/*
 * Copyright 2009, Mikhail Panasyuk, otoko@yandex.ru.
 *
 * Copyright 2008, Ingo Weinhold, ingo_weinhold@gmx.de.
 * Copyright 2002-2007, Axel Dörfler, axeld@pinc-software.de.
 * Copyright 2002, Angelo Mottola, a.mottola@libero.it.
 * Distributed under the terms of the MIT License.
 *
 * Copyright 2001-2002, Travis Geiselbrecht. All rights reserved.
 * Distributed under the terms of the NewOS License.
 */

/*! The thread scheduler */


#include <OS.h>

#include <cpu.h>
#include <int.h>
#include <kernel.h>
#include <kscheduler.h>
#include <listeners.h>
#include <scheduler_defs.h>
#include <thread.h>
#include <timer.h>
#include <user_debugger.h>

#include "scheduler_tracing.h"


//#define TRACE_SCHEDULER
#ifdef TRACE_SCHEDULER
#	define TRACE(x) dprintf x
#else
#	define TRACE(x) ;
#endif


// The run queue. Holds the threads ready to run ordered by priority.
static struct thread *sRunQueue = NULL;


// #pragma mark - Optimisation tricks


// Run queue support structures, while complicating things a bit allows fast
// search and other cool things.

typedef struct priority_info {
	int16			threads;	// number of threads in subqueue
	int16			prev;		// previous priority number with threads != 0
	int16			next;		// next priority number with threads != 0
	struct thread*	first;		// first thread in subqueue of threads
	struct thread*	last;		// last thread in subqueue of threads
} priority_info;


/*typedef struct queue_mask {
	queue_mask(void)
	{
		Init();
	}
	
	void Init()
	{
		fData[0] = fData[1] = fData[2] = fData[3] = 0;
	}

	inline void SetBit(int16 bit, bool value) {
		int16 dataElement = bit / 32;
		int16 elementBit = bit % 32;
		uint32 mask = 0x00000001 << elementBit;
		if (value) {
			fData[dataElement] |= mask;
		} else {
			fData[dataElement] &= 0xffffffff ^ mask;
		}
	}

	// host byte order type doesn't matter because structure is used only for
	// comparations as whole

	inline bool operator== (const queue_mask & qm) const
	{
		if (fData[0] == qm.fData[0]
			&& fData[1] == qm.fData[1]
			&& fData[2] == qm.fData[2]
			&& fData[3] == qm.fData[3]
		)
			return true;

		return false;
	}

	inline bool operator!= (const queue_mask & qm) const
	{
		if (fData[0] != qm.fData[0]
			|| fData[1] != qm.fData[1]
			|| fData[2] != qm.fData[2]
			|| fData[3] != qm.fData[3]
		)
			return true;

		return false;
	}
	
	uint32 fData[4];
} queue_mask;


typedef struct queue_status {
	queue_status(void)
	{
		Reset();
	}
	
	inline void Reset()
	{
		fHasChanges = false;
		fManyChanges = false;
	}

	inline void CommitChange(int16 value, bool isAdd)
	{
		if (fManyChanges)
			return;

		if (fHasChanges) {
			if (fLastValue == value && fLastWasAdd != isAdd)
				fHasChanges = false;
			else
				fManyChanges = true;
		} else {
			fLastWasAdd = isAdd;
			fLastValue = value;
			fHasChanges = true;
		}
	}

	int16	fLastValue;
	bool	fLastWasAdd;
	bool	fHasChanges;
	bool	fManyChanges;
} queue_status;*/


#define QUEUE_INFO_SIZE		(B_REAL_TIME_PRIORITY + 1)

static priority_info gRunQueueInfo[QUEUE_INFO_SIZE];
//static queue_mask gRunQueueMask;
//static queue_status gRunQueueStatus;
//static int16 gRunQueueDepth = 0;


static void
reset_queue_infos()
{
	for (int i = 0; i < QUEUE_INFO_SIZE; i++)
	{
		gRunQueueInfo[i].threads = 0;
		gRunQueueInfo[i].prev = -1;
		gRunQueueInfo[i].next = -1;
		gRunQueueInfo[i].first = NULL;
		gRunQueueInfo[i].last = NULL;
	}
}


// #pragma mark -


static int
_rand(void)
{
	static int next = 0;

	if (next == 0)
		next = system_time();

	next = next * 1103515245 + 12345;
	return (next >> 16) & 0x7FFF;
}


static int
dump_run_queue(int argc, char **argv)
{
	struct thread *thread;

	thread = sRunQueue;
	if (!thread)
		kprintf("Run queue is empty!\n");
	else {
		kprintf("thread    id      priority name\n");
		while (thread) {
			kprintf("%p  %-7ld %-8ld %s\n", thread, thread->id,
				thread->priority, thread->name);
			thread = thread->queue_next;
		}
	}

	return 0;
}


/*!	Enqueues the thread into the run queue.
	Note: thread lock must be held when entering this function
*/
static bool
simple_enqueue_in_run_queue(struct thread *thread)
{
	thread->state = thread->next_state = B_THREAD_READY;

	// find thread in run queue
	int16 priority = thread->next_priority;
	priority_info *info = &(gRunQueueInfo[priority]);

	struct thread *next, *prev;
	if (info->threads != 0)
	{
		prev = info->last;
		if (info->next != -1) {
			next = gRunQueueInfo[info->next].first;
		} else {
			next = NULL;
		}
	} else {
		prev = NULL;
		next = sRunQueue;
		if (next)
		{
			int16 nextPriority = next->priority;
			while(nextPriority != -1)
			{
				if (nextPriority > priority) {
					prev = gRunQueueInfo[nextPriority].last;
					nextPriority = gRunQueueInfo[nextPriority].next;
				} else {
					next = gRunQueueInfo[nextPriority].first;
					break;
				}
			}
			if (next->priority > priority) {
				next = NULL;
			}
		}
	}

	T(EnqueueThread(thread, prev, next));

	thread->queue_next = next;
	if (prev)
		prev->queue_next = thread;
	else
		sRunQueue = thread;

	if (info->threads == 0)
	{
//		gRunQueueDepth++;
//		gRunQueueMask.SetBit(priority, true);
		
		if (prev) {
			gRunQueueInfo[prev->priority].next = priority;
			info->prev = prev->priority;
		}
		
		if (next) {
			info->next = next->priority;
			gRunQueueInfo[next->priority].prev = priority;
		}

		info->first = thread;
	}
	info->last = thread;
	info->threads++;

//	gRunQueueStatus.CommitChange(priority, true);

	thread->priority = thread->next_priority;

	// notify listeners
	NotifySchedulerListeners(&SchedulerListener::ThreadEnqueuedInRunQueue,
		thread);

	return thread->priority > thread_get_current_thread()->priority;
}


static void
remove_from_run_queue(struct thread *thread)
{
	// find thread in subqueue
	int16 priority = thread->priority;
	priority_info *info = &(gRunQueueInfo[priority]);
	
	struct thread *prev = NULL;
	if (info->prev != -1) {
		prev = gRunQueueInfo[info->prev].last;
	}

	struct thread *item;
	for (item = info->first; item && item != thread;
			item = item->queue_next) {
		if (prev)
			prev = prev->queue_next;
		else
			prev = info->first;
	}

	ASSERT(item == thread);

	struct thread *next = thread->queue_next;
	if (prev)
		prev->queue_next = next;
	else
		sRunQueue = next;

	info->threads--;
	if (info->threads == 0)
	{
//		gRunQueueDepth--;
//		gRunQueueMask.SetBit(priority, false);
	
		if (prev) {
			if (prev->queue_next)
				gRunQueueInfo[prev->priority].next = prev->queue_next->priority;
			else
				gRunQueueInfo[prev->priority].next = -1;
		}
		
		if (next) {
			if (prev)
				gRunQueueInfo[next->priority].prev = prev->priority;
			else
				gRunQueueInfo[next->priority].prev = -1;
		}
		
		// reset thread priority info
		info->prev = -1;
		info->next = -1;
		info->first = NULL;
		info->last = NULL;
	} else {
		if (info->first == thread)
			info->first = next;
		if (info->last == thread)
			info->last = prev;
	}

//	gRunQueueStatus.CommitChange(priority, false);
}


/*
	Note: removes only from head of subqueue.
*/
static void
fast_remove_from_run_queue(struct thread *thread)
{
	// find thread in run queue
	int16 priority = thread->priority;
	priority_info *info = &(gRunQueueInfo[priority]);
	
	ASSERT(thread == info->first);

	struct thread *prev = NULL;
	if (info->prev != -1) {
		prev = gRunQueueInfo[info->prev].last;
	}

	struct thread *next = thread->queue_next;
	if (prev)
		prev->queue_next = next;
	else
		sRunQueue = next;

	info->threads--;
	if (info->threads == 0)
	{
//		gRunQueueDepth--;
//		gRunQueueMask.SetBit(priority, false);
	
		if (prev) {
			if (prev->queue_next)
				gRunQueueInfo[prev->priority].next = prev->queue_next->priority;
			else
				gRunQueueInfo[prev->priority].next = -1;
		}
		
		if (next) {
			if (prev)
				gRunQueueInfo[next->priority].prev = prev->priority;
			else
				gRunQueueInfo[next->priority].prev = -1;
		}
		
		// reset thread priority info
		info->prev = -1;
		info->next = -1;
		info->first = NULL;
		info->last = NULL;
	} else {
		// again, removing always from head of subqueue
		info->first = next;
	}

//	gRunQueueStatus.CommitChange(priority, false);
}


/*!	Sets the priority of a thread.
	Note: thread lock must be held when entering this function
*/
static void
simple_set_thread_priority(struct thread *thread, int32 priority)
{
	if (priority == thread->priority)
		return;

	if (thread->state != B_THREAD_READY) {
		thread->priority = thread->next_priority = priority;
		return;
	}

	// The thread is in the run queue. We need to remove it and re-insert it at
	// a new position.

	T(RemoveThread(thread));

	// notify listeners
	NotifySchedulerListeners(&SchedulerListener::ThreadRemovedFromRunQueue,
		thread);

	// remove the thread
	remove_from_run_queue(thread);

	// set priority and re-insert
	thread->priority = thread->next_priority = priority;
	simple_enqueue_in_run_queue(thread);
}


static void
context_switch(struct thread *fromThread, struct thread *toThread)
{
	if ((fromThread->flags & THREAD_FLAGS_DEBUGGER_INSTALLED) != 0)
		user_debug_thread_unscheduled(fromThread);

	toThread->previous_cpu = toThread->cpu = fromThread->cpu;
	fromThread->cpu = NULL;

	arch_thread_set_current_thread(toThread);
	arch_thread_context_switch(fromThread, toThread);

	// Looks weird, but is correct. fromThread had been unscheduled earlier,
	// but is back now. The notification for a thread scheduled the first time
	// happens in thread.cpp:thread_kthread_entry().
	if ((fromThread->flags & THREAD_FLAGS_DEBUGGER_INSTALLED) != 0)
		user_debug_thread_scheduled(fromThread);
}


static int32
reschedule_event(timer *unused)
{
	if (thread_get_current_thread()->keep_scheduled > 0)
		return B_HANDLED_INTERRUPT;

	// this function is called as a result of the timer event set by the
	// scheduler returning this causes a reschedule on the timer event
	thread_get_current_thread()->cpu->preempted = 1;
	return B_INVOKE_SCHEDULER;
}


// #pragma mark - Haiku-like algorithm


static struct thread*
select_next_thread_to_run()
{
	struct thread *nextThread = sRunQueue;

	// in case we don't have any non-idle threads to run
	if (nextThread->priority == B_IDLE_PRIORITY)
		return nextThread;

	// select only from real-time threads if any
	if (nextThread->priority >= B_FIRST_REAL_TIME_PRIORITY)
		return nextThread;

	// select from normal threads

	// walk through run queue
	
	int16 nextLowerPriority;
	while (1) {
		nextLowerPriority = gRunQueueInfo[nextThread->priority].next;

		// never skip last non-idle subqueue of threads
		if (nextLowerPriority == B_IDLE_PRIORITY)
			break;

		// skip subqueue of threads sometimes
		if (_rand() > 0x1a00) break;

		nextThread = gRunQueueInfo[nextLowerPriority].first;
	}
	return nextThread;
}


// #pragma mark -


/*!	Runs the scheduler.
	Note: expects thread spinlock to be held
*/
static void
simple_reschedule(void)
{
	struct thread *oldThread = thread_get_current_thread();
	struct thread *nextThread;//, *prevThread;

	TRACE(("reschedule(): current thread = %ld\n", oldThread->id));

	oldThread->cpu->invoke_scheduler = false;

	oldThread->state = oldThread->next_state;
	switch (oldThread->next_state) {
		case B_THREAD_RUNNING:
		case B_THREAD_READY:
			TRACE(("enqueueing thread %ld into run queue priority = %ld\n",
				oldThread->id, oldThread->priority));
			simple_enqueue_in_run_queue(oldThread);
			break;
		case B_THREAD_SUSPENDED:
			TRACE(("reschedule(): suspending thread %ld\n", oldThread->id));
			break;
		case THREAD_STATE_FREE_ON_RESCHED:
			break;
		default:
			TRACE(("not enqueueing thread %ld into run queue next_state = %ld\n",
				oldThread->id, oldThread->next_state));
			break;
	}

// ToDo: implement 'if 0'-behaviour if needed
//       (ignore threads that called thread_yield() once)

/*	nextThread = sRunQueue;
	prevThread = NULL;

	while (nextThread) {
		// select next thread from the run queue
		while (nextThread && nextThread->priority > B_IDLE_PRIORITY) {
#if 0
			if (oldThread == nextThread && nextThread->was_yielded) {
				// ignore threads that called thread_yield() once
				nextThread->was_yielded = false;
				prevThread = nextThread;
				nextThread = nextThread->queue_next;
			}
#endif

			// always extract real time threads
			if (nextThread->priority >= B_FIRST_REAL_TIME_PRIORITY)
				break;

			// never skip last non-idle normal thread
			if (nextThread->queue_next
				&& nextThread->queue_next->priority == B_IDLE_PRIORITY)
				break;

			// skip normal threads sometimes (roughly 20%)
			if (_rand() > 0x1a00)
				break;

			// skip until next lower priority
			int32 priority = nextThread->priority;
			do {
				prevThread = nextThread;
				nextThread = nextThread->queue_next;
			} while (nextThread->queue_next != NULL
				&& priority == nextThread->queue_next->priority
				&& nextThread->queue_next->priority > B_IDLE_PRIORITY);
		}

		break;
	}
	
	if (!nextThread)
		panic("reschedule(): run queue is empty!\n");*/

	if (!sRunQueue)
		panic("reschedule(): run queue is empty!\n");

	nextThread = select_next_thread_to_run();

	// extract selected thread from the run queue
	fast_remove_from_run_queue(nextThread);

	T(ScheduleThread(nextThread, oldThread));

	// notify listeners
	NotifySchedulerListeners(&SchedulerListener::ThreadScheduled,
		oldThread, nextThread);

	nextThread->state = B_THREAD_RUNNING;
	nextThread->next_state = B_THREAD_READY;
	oldThread->was_yielded = false;

	// track kernel time (user time is tracked in thread_at_kernel_entry())
	bigtime_t now = system_time();
	oldThread->kernel_time += now - oldThread->last_time;
	nextThread->last_time = now;

	// track CPU activity
	if (!thread_is_idle_thread(oldThread)) {
		oldThread->cpu->active_time +=
			(oldThread->kernel_time - oldThread->cpu->last_kernel_time)
			+ (oldThread->user_time - oldThread->cpu->last_user_time);
	}

	if (!thread_is_idle_thread(nextThread)) {
		oldThread->cpu->last_kernel_time = nextThread->kernel_time;
		oldThread->cpu->last_user_time = nextThread->user_time;
	}

	if (nextThread != oldThread || oldThread->cpu->preempted) {
		bigtime_t quantum = 3000;	// ToDo: calculate quantum!
		timer *quantumTimer = &oldThread->cpu->quantum_timer;

		if (!oldThread->cpu->preempted)
			cancel_timer(quantumTimer);

		oldThread->cpu->preempted = 0;
		add_timer(quantumTimer, &reschedule_event, quantum,
			B_ONE_SHOT_RELATIVE_TIMER | B_TIMER_ACQUIRE_THREAD_LOCK);

		if (nextThread != oldThread)
			context_switch(oldThread, nextThread);
	}
}


static void
simple_on_thread_create(struct thread* thread)
{
	// do nothing
}


static void
simple_on_thread_init(struct thread* thread)
{
	// do nothing
}


static void
simple_on_thread_destroy(struct thread* thread)
{
	// do nothing
}


/*!	This starts the scheduler. Must be run in the context of the initial idle
	thread. Interrupts must be disabled and will be disabled when returning.
*/
static void
simple_start(void)
{
	GRAB_THREAD_LOCK();

	simple_reschedule();

	RELEASE_THREAD_LOCK();
}


static scheduler_ops kSimpleOps = {
	simple_enqueue_in_run_queue,
	simple_reschedule,
	simple_set_thread_priority,
	simple_on_thread_create,
	simple_on_thread_init,
	simple_on_thread_destroy,
	simple_start
};


// #pragma mark -


void
scheduler_simple_init()
{
	gScheduler = &kSimpleOps;
	reset_queue_infos();
	add_debugger_command_etc("run_queue", &dump_run_queue,
		"List threads in run queue", "\nLists threads in run queue", 0);
}
